package admin

import (
	"database/sql"
	"fmt"
	"net/http"
	"slices"
	"strconv"
	"strings"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *adminCtr) GetGroups(ctx echo.Context) error {
	htmxReq := ctx.Request().Header.Get("HX-Request")
	page := controller.NewPage(ctx, c.Container)
	page.Pager = controller.NewPager(ctx, 50)
	page.Title = "Admin - Groups"

	var groups []sqlc.Group
	var err error
	query := ctx.QueryParam("search")
	if query != "" {
		dbQuery := "%" + query + "%"
		groups, err = c.Container.DB.C.SearchGroups(ctx.Request().Context(), dbQuery)
	} else {
		groups, err = c.Container.DB.C.GetGroups(ctx.Request().Context())
	}

	if err != nil {
		return c.Fail(err, "unable to retrieve groups")
	}

	form := views.AdminAddGroupForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.AdminAddGroupForm)
	}

	v := views.AdminGroups{
		Groups:        groups,
		GroupListOnly: htmxReq == "true",
		Form:          form,
		Page:          page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostGroups(ctx echo.Context) error {
	var form views.AdminAddGroupForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetGroups(ctx)
	}

	groupID, err := c.Container.DB.C.CreateGroup(ctx.Request().Context(), form.Name)
	if err != nil {
		return c.Fail(err, "unable to create group")
	}

	return c.Redirect(ctx, pages.Names.Admin.GroupDetails, groupID)
}

func (c *adminCtr) GetGroupDetails(ctx echo.Context) error {
	htmxReq := ctx.Request().Header.Get("HX-Request")

	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	group, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested group")
	}

	members, err := c.Container.DB.C.GetGroupMembers(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested group members")
	}

	query := ctx.QueryParam("search")
	lowerQuery := strings.ToLower(query)
	if query != "" {
		// Filter the members using the search query
		members = slices.DeleteFunc(members, func(u sqlc.User) bool {
			return !strings.Contains(strings.ToLower(u.Email), lowerQuery) &&
				!strings.Contains(strings.ToLower(u.Name.String), lowerQuery)
		})
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Admin - Group"

	form := views.AdminGroupDetailsForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.AdminGroupDetailsForm)
	}

	v := views.AdminGroupDetails{
		Group:         group,
		Members:       members,
		Form:          form,
		GroupListOnly: htmxReq == "true",
		Page:          page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostGroupDetails(ctx echo.Context) error {
	var form views.AdminGroupDetailsForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetGroupDetails(ctx)
	}

	// Check that there is a user for this email.
	email := strings.ToLower(form.AddUserEmail)
	userToAdd, err := c.Container.DB.C.GetUserByEmail(ctx.Request().Context(), email)
	if err == sql.ErrNoRows {
		form.Submission.SetFieldError("AddUserEmail", "A user for that email does not exist")
		return c.GetGroupDetails(ctx)
	}
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	group, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested group")
	}

	groupMembers, err := c.Container.DB.C.GetGroupMembers(ctx.Request().Context(), group.ID)
	if err != nil {
		return c.Fail(err, "unable to retrieve group members")
	}

	alreadyMember := slices.ContainsFunc(groupMembers, func(u sqlc.User) bool {
		return u.ID == userToAdd.ID
	})
	if alreadyMember {
		form.Submission.SetFieldError("AddUserEmail", "User is already a member of this group")
	}

	if form.Submission.HasErrors() {
		return c.GetGroupDetails(ctx)
	}

	err = c.Container.DB.C.AddUserToGroup(ctx.Request().Context(), sqlc.AddUserToGroupParams{
		UserID:  userToAdd.ID,
		GroupID: group.ID,
	})
	if err != nil {
		return c.Fail(err, "unable to add user to group")
	}

	return c.Redirect(ctx, pages.Names.Admin.GroupDetails, groupID)
}

func (c *adminCtr) GetDeleteGroup(ctx echo.Context) error {
	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	group, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested group")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Delete Group"

	v := views.ActionConfirmation{
		Heading:     "Delete group",
		Prompt:      fmt.Sprintf("Are you sure you want to delete the group %s? This action cannot be undone.", group.Name),
		ButtonText:  "Delete",
		Destructive: true,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostDeleteGroup(ctx echo.Context) error {
	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetDeleteGroup(ctx)
	}

	// NOTE: Follows foreign keys to clean up memberships for this group too.
	err = c.Container.DB.C.DeleteGroup(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to delete the group")
	}

	return ctx.Redirect(http.StatusFound, ctx.Echo().Reverse(pages.Names.Admin.Groups))
}

func (c *adminCtr) GetRemoveGroupUser(ctx echo.Context) error {
	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	group, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), groupID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested group")
	}

	userID, err := strconv.Atoi(ctx.Param("user_id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	user, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), userID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested user")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Remove user"

	v := views.ActionConfirmation{
		Heading:     "Remove user",
		Prompt:      fmt.Sprintf("Are you sure you want to remove %s from group %s?", user.Email, group.Name),
		ButtonText:  "Remove",
		Destructive: true,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostRemoveGroupUser(ctx echo.Context) error {
	groupID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested group")
	}

	userID, err := strconv.Atoi(ctx.Param("user_id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetRemoveGroupUser(ctx)
	}

	err = c.Container.DB.C.RemoveUserFromGroup(ctx.Request().Context(), sqlc.RemoveUserFromGroupParams{
		UserID:  userID,
		GroupID: groupID,
	})
	if err != nil {
		return c.Fail(err, "unable to remove user from group")
	}

	return ctx.Redirect(http.StatusFound, ctx.Echo().Reverse(pages.Names.Admin.GroupDetails, groupID))
}
