package admin

import (
	"fmt"
	"net/http"
	"strconv"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *adminCtr) GetTags(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Title = "Tags"

	tags, err := c.Container.DB.C.GetTags(ctx.Request().Context())
	if err != nil {
		return c.Fail(err, "unable to retrieve tags")
	}

	groups, err := c.Container.DB.C.GetGroups(ctx.Request().Context())
	if err != nil {
		return c.Fail(err, "unable to retrieve groups")
	}

	form := views.AdminAddTagForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.AdminAddTagForm)
	}

	v := views.AdminTags{
		Tags:   tags,
		Groups: groups,
		Form:   form,
		Page:   page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostTags(ctx echo.Context) error {
	var form views.AdminAddTagForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	groupID := conv.NullInt{Valid: false}
	if form.RestrictedGroupID != "" {
		parsedID, err := strconv.Atoi(form.RestrictedGroupID)
		if err != nil {
			form.Submission.SetFieldError("RestrictedGroupID", "Invalid group")
		} else {
			_, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), parsedID)
			if err != nil {
				form.Submission.SetFieldError("RestrictedGroupID", "Invalid group")
			}
			groupID = conv.NullInt{Valid: true, Int: parsedID}
		}
	}

	if form.Submission.HasErrors() {
		return c.GetTags(ctx)
	}

	tagID, err := c.Container.DB.C.CreateTag(ctx.Request().Context(), sqlc.CreateTagParams{
		Name:              form.Name,
		RestrictedGroupID: groupID,
	})
	if err != nil {
		return c.Fail(err, "unable to create tag")
	}

	return c.Redirect(ctx, pages.Names.Admin.TagDetails, tagID)
}

func (c *adminCtr) GetTagDetails(ctx echo.Context) error {
	tagID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested tag")
	}

	tag, err := c.Container.DB.C.GetTagByID(ctx.Request().Context(), tagID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested tag")
	}

	var restrictedGroup *sqlc.Group
	if tag.RestrictedGroupID.Valid {
		group, err := c.Container.DB.C.GetGroupByID(ctx.Request().Context(), tag.RestrictedGroupID.Int)
		if err != nil {
			return c.Fail(err, "unable to retrieve requested tag")
		}
		restrictedGroup = &group
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Admin - Tag"

	v := views.AdminTagDetails{
		Tag:             tag,
		RestrictedGroup: restrictedGroup,
		Page:            page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) GetDeleteTag(ctx echo.Context) error {
	tagID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested tag")
	}

	tag, err := c.Container.DB.C.GetTagByID(ctx.Request().Context(), tagID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested tag")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Delete Tag"

	v := views.ActionConfirmation{
		Heading:     "Delete tag",
		Prompt:      fmt.Sprintf("Are you sure you want to delete the tag %s? This action cannot be undone.", tag.Name),
		ButtonText:  "Delete",
		Destructive: true,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostDeleteTag(ctx echo.Context) error {
	tagID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested tag")
	}

	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetDeleteGroup(ctx)
	}

	// NOTE: Follows foreign keys.
	err = c.Container.DB.C.DeleteTag(ctx.Request().Context(), tagID)
	if err != nil {
		return c.Fail(err, "unable to delete the tag")
	}

	return ctx.Redirect(http.StatusFound, ctx.Echo().Reverse(pages.Names.Admin.Tags))
}
