package event

import (
	"errors"
	"fmt"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type create struct {
	controller.Controller
}

func (c *create) Get(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Title = "New Event"
	page.Metatags.Description = fmt.Sprintf("Create new event on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	form := views.EventCreateForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.EventCreateForm)
	}

	tags, err := c.Container.DB.C.GetTagsAllowedForUser(ctx.Request().Context(), page.AuthUserGroups)
	if err != nil {
		logger := context.GetLogger(ctx)
		logger.Error("unable to get allowed tags for user",
			zap.Error(err))
	}

	allowedTags := generic.Map(tags, func(t sqlc.Tag) string {
		return t.Name
	})

	v := views.EventCreate{
		Location:    c.Container.Config.App.Loc,
		AllowedTags: allowedTags,
		ActionHref:  ctx.Echo().Reverse(pages.Names.Event.New),
		CancelHref:  ctx.Echo().Reverse(pages.Names.Event.Browse),
		SubmitText:  "Create",
		Form:        form,
		Page:        page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *create) Post(ctx echo.Context) error {
	var form views.EventCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	form.Process(context.GetLogger(ctx), c.Container.Config.App.Loc)

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user"), "unable to lookup authenticated user")
	}
	groups, ok := ctx.Get(context.AuthenticatedUserGroupsKey).([]string)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user groups"), "unable to lookup authenticated user groups")
	}

	allowedTags, err := c.Container.DB.C.GetTagsAllowedForUser(ctx.Request().Context(), groups)
	if err != nil {
		return c.Fail(err, "unable to get allowed tags for user")
	}

	form.Sanitize(c.Container.Sanitizer)

	eventID, err := c.Container.DB.C.CreateEvent(ctx.Request().Context(), sqlc.CreateEventParams{
		Title:       form.Title,
		Description: form.Description,
		Creator:     user.ID,
		Location:    form.Location,
		HasTime:     form.StartTime != "",
		StartTime:   form.ProcessedStartTime,
		EndTime:     form.ProcessedEndTime,
	})
	if err != nil {
		return c.Fail(err, "unable to save a new event")
	}

	c.Container.DB.C.IncrementEventsCreatedCount(ctx.Request().Context(), user.ID)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment posts created stat",
			zap.Error(err))
	}

	// Process the tags. Only add tags that the user is allowed to add.
	allowedTagsMap := make(map[string]sqlc.Tag)
	for _, t := range allowedTags {
		allowedTagsMap[t.Name] = t
	}

	for _, t := range form.ProcessedTags {
		tag, ok := allowedTagsMap[t]
		if !ok {
			msg.Danger(ctx, "You do not have permission to use the tag "+t)
			continue
		}

		err = c.Container.DB.C.AddEventTag(ctx.Request().Context(), sqlc.AddEventTagParams{
			TagID:   tag.ID,
			EventID: int(eventID),
		})
		if err != nil {
			msg.Danger(ctx, "Unable to add tag "+t+" to event")
			continue
		}
	}

	return c.Redirect(ctx, pages.Names.Event.One, eventID)
}
