package event

import (
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *update) EditGet(ctx echo.Context) error {
	event, ok := ctx.Get(context.EventKey).(sqlc.Event)
	if !ok {
		context.GetLogger(ctx).Warn("Unable to retrieve event from context unexpectedly")
		return echo.ErrInternalServerError
	}

	eventTags, err := c.Container.DB.C.GetEventTags(ctx.Request().Context(), event.ID)
	if err != nil {
		context.GetLogger(ctx).Error("Unable to retrieve event tags", zap.Error(err))
		return echo.ErrInternalServerError
	}
	tags := generic.Map(eventTags, func(t sqlc.Tag) string {
		return t.Name
	})
	tagsStr := strings.Join(tags, ", ")

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Edit Event"
	page.Metatags.Description = fmt.Sprintf("Edit event on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	allowedTags, err := c.Container.DB.C.GetTagsAllowedForUser(ctx.Request().Context(), page.AuthUserGroups)
	if err != nil {
		logger := context.GetLogger(ctx)
		logger.Error("unable to get allowed tags for user",
			zap.Error(err))
	}

	allowedTagsNames := generic.Map(allowedTags, func(t sqlc.Tag) string {
		return t.Name
	})

	var form views.EventCreateForm
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.EventCreateForm)
	} else {
		startTime := ""
		if event.HasTime {
			startTime = event.StartTime.Format("15:04")
		}

		form = views.EventCreateForm{
			Title:       event.Title,
			Description: event.Description,
			Location:    event.Location,
			StartTime:   startTime,
			Date:        event.StartTime.Format(time.DateOnly),
			Tags:        tagsStr,
		}
		if event.EndTime.Valid {
			form.EndTime = event.EndTime.Time.Format("15:04")
		}
	}

	v := views.EventCreate{
		Location:    c.Container.Config.App.Loc,
		AllowedTags: allowedTagsNames,
		ActionHref:  ctx.Echo().Reverse(pages.Names.Event.Edit, event.ID),
		CancelHref:  ctx.Echo().Reverse(pages.Names.Event.One, event.ID),
		SubmitText:  "Update",
		Form:        form,
		Page:        page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *update) EditPost(ctx echo.Context) error {
	eventID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		msg.Warning(ctx, "The requested event could not be found.")
		return c.Redirect(ctx, pages.Names.Event.Browse)
	}

	var form views.EventCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	form.Process(context.GetLogger(ctx), c.Container.Config.App.Loc)

	if form.Submission.HasErrors() {
		return c.EditGet(ctx)
	}

	form.Sanitize(c.Container.Sanitizer)

	err = c.Container.DB.C.UpdateEvent(ctx.Request().Context(), sqlc.UpdateEventParams{
		Title:       form.Title,
		Description: form.Description,
		Location:    form.Location,
		HasTime:     form.StartTime != "",
		StartTime:   form.ProcessedStartTime, // Valid is checked above
		EndTime:     form.ProcessedEndTime,
		ID:          eventID,
	})
	if err != nil {
		return c.Fail(err, "unable to update event")
	}

	existingTags, err := c.Container.DB.C.GetEventTags(ctx.Request().Context(), eventID)
	if err != nil {
		context.GetLogger(ctx).Error("Unable to retrieve existing event tags while editing", zap.Error(err))
	}

	existingTagsMap := make(map[string]sqlc.Tag)
	for _, t := range existingTags {
		existingTagsMap[t.Name] = t
	}

	groups, ok := ctx.Get(context.AuthenticatedUserGroupsKey).([]string)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user groups"), "unable to lookup authenticated user groups")
	}

	allowedTags, err := c.Container.DB.C.GetTagsAllowedForUser(ctx.Request().Context(), groups)
	if err != nil {
		return c.Fail(err, "unable to get allowed tags for user")
	}

	// Process the tags. Only add tags that the user is allowed to add.
	allowedTagsMap := make(map[string]sqlc.Tag)
	for _, t := range allowedTags {
		allowedTagsMap[t.Name] = t
	}

	for _, t := range form.ProcessedTags {
		existingTag, existing := existingTagsMap[t]
		allowedTag, allowed := allowedTagsMap[t]

		if existing {
			delete(existingTagsMap, t)
		}

		if !existing && allowed {
			err = c.Container.DB.C.AddEventTag(ctx.Request().Context(), sqlc.AddEventTagParams{
				EventID: eventID,
				TagID:   allowedTag.ID,
			})
			if err != nil {
				context.GetLogger(ctx).Error("Unable to add tag "+t+" to event", zap.Error(err))
			}
		} else if existing && !allowed {
			err = c.Container.DB.C.RemoveEventTag(ctx.Request().Context(), sqlc.RemoveEventTagParams{
				EventID: eventID,
				TagID:   existingTag.ID,
			})
			if err != nil {
				context.GetLogger(ctx).Error("Unable to remove tag "+t+" from event", zap.Error(err))
			}
		}
	}

	for _, t := range existingTagsMap {
		err = c.Container.DB.C.RemoveEventTag(ctx.Request().Context(), sqlc.RemoveEventTagParams{
			EventID: eventID,
			TagID:   t.ID,
		})
		if err != nil {
			context.GetLogger(ctx).Error("Unable to remove tag "+t.Name+" from event", zap.Error(err))
		}
	}

	return c.Redirect(ctx, pages.Names.Event.One, eventID)
}
